@@ -583,9 +583,12 @@ static bool eap_tls_tunnel_init(struct eap_state *eap)
 		return false;
 	}
 
-	if (getenv("IWD_TLS_DEBUG"))
+	if (getenv("IWD_TLS_DEBUG")) {
 		l_tls_set_debug(eap_tls->tunnel, eap_tls_tunnel_debug, eap,
 									NULL);
+		l_tls_set_cert_dump_path(eap_tls->tunnel,
+					"/tmp/iwd-tls-debug-server-cert.pem");
+	}
 
 	if (eap_tls->client_cert || eap_tls->client_key) {
 		if (!l_tls_set_auth_data(eap_tls->tunnel, eap_tls->client_cert,
