@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -31,6 +31,7 @@
 #include <ell/ell.h>
 
 #include "src/missing.h"
+#include "src/module.h"
 #include "src/crypto.h"
 #include "src/eapol.h"
 #include "src/ie.h"
@@ -40,18 +41,17 @@
 #include "src/handshake.h"
 #include "src/watchlist.h"
 #include "src/erp.h"
+#include "src/iwd.h"
 
-struct l_queue *state_machines;
-struct l_queue *preauths;
-struct watchlist frame_watches;
+static struct l_queue *state_machines;
+static struct l_queue *preauths;
+static struct watchlist frame_watches;
 static uint32_t eapol_4way_handshake_time = 2;
 
-eapol_rekey_offload_func_t rekey_offload = NULL;
+static eapol_rekey_offload_func_t rekey_offload = NULL;
 
-eapol_tx_packet_func_t tx_packet = NULL;
-void *tx_user_data;
-
-uint32_t next_frame_watch_id;
+static eapol_tx_packet_func_t tx_packet = NULL;
+static void *tx_user_data;
 
 #define VERIFY_IS_ZERO(field)						\
 	do {								\
@@ -104,6 +104,7 @@ bool eapol_calculate_mic(enum ie_rsn_akm_suite akm, const uint8_t *kck,
 		switch (akm) {
 		case IE_RSN_AKM_SUITE_SAE_SHA256:
 		case IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256:
+		case IE_RSN_AKM_SUITE_OSEN:
 			return cmac_aes(kck, 16, frame, frame_len,
 						mic, mic_len);
 		case IE_RSN_AKM_SUITE_OWE:
@@ -157,6 +158,7 @@ bool eapol_verify_mic(enum ie_rsn_akm_suite akm, const uint8_t *kck,
 		switch (akm) {
 		case IE_RSN_AKM_SUITE_SAE_SHA256:
 		case IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256:
+		case IE_RSN_AKM_SUITE_OSEN:
 			checksum = l_checksum_new_cmac_aes(kck, 16);
 			break;
 		case IE_RSN_AKM_SUITE_OWE:
@@ -257,6 +259,7 @@ uint8_t *eapol_decrypt_key_data(enum ie_rsn_akm_suite akm, const uint8_t *kek,
 		case IE_RSN_AKM_SUITE_SAE_SHA256:
 		case IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256:
 		case IE_RSN_AKM_SUITE_OWE:
+		case IE_RSN_AKM_SUITE_OSEN:
 			if (key_data_len < 24 || key_data_len % 8)
 				return NULL;
 
@@ -276,7 +279,7 @@ uint8_t *eapol_decrypt_key_data(enum ie_rsn_akm_suite akm, const uint8_t *kek,
 		break;
 	default:
 		return NULL;
-	};
+	}
 
 	buf = l_new(uint8_t, expected_len);
 
@@ -821,7 +824,6 @@ struct eapol_sm {
 	struct handshake_state *handshake;
 	enum eapol_protocol_version protocol_version;
 	uint64_t replay_counter;
-	eapol_sm_event_func_t event_func;
 	void *user_data;
 	struct l_timeout *timeout;
 	struct l_timeout *eapol_start_timeout;
@@ -832,8 +834,10 @@ struct eapol_sm {
 	bool use_eapol_start:1;
 	bool require_handshake:1;
 	bool eap_exchanged:1;
+	bool last_eap_unencrypted:1;
 	struct eap_state *eap;
 	struct eapol_frame *early_frame;
+	bool early_frame_unencrypted : 1;
 	uint32_t watch_id;
 	uint8_t installed_gtk_len;
 	uint8_t installed_gtk[CRYPTO_MAX_GTK_LEN];
@@ -862,8 +866,6 @@ static void eapol_sm_destroy(void *value)
 	explicit_bzero(sm->installed_igtk, sizeof(sm->installed_igtk));
 
 	l_free(sm);
-
-	l_queue_remove(state_machines, sm);
 }
 
 struct eapol_sm *eapol_sm_new(struct handshake_state *hs)
@@ -874,7 +876,7 @@ struct eapol_sm *eapol_sm_new(struct handshake_state *hs)
 
 	sm->handshake = hs;
 
-	if (hs->settings_8021x)
+	if (hs->settings_8021x && !hs->authenticator)
 		sm->use_eapol_start = true;
 
 	sm->require_handshake = true;
@@ -884,13 +886,9 @@ struct eapol_sm *eapol_sm_new(struct handshake_state *hs)
 
 void eapol_sm_free(struct eapol_sm *sm)
 {
-	eapol_sm_destroy(sm);
-}
+	l_queue_remove(state_machines, sm);
 
-void eapol_sm_set_protocol_version(struct eapol_sm *sm,
-				enum eapol_protocol_version protocol_version)
-{
-	sm->protocol_version = protocol_version;
+	eapol_sm_destroy(sm);
 }
 
 void eapol_sm_set_listen_interval(struct eapol_sm *sm, uint16_t interval)
@@ -903,21 +901,19 @@ void eapol_sm_set_user_data(struct eapol_sm *sm, void *user_data)
 	sm->user_data = user_data;
 }
 
-void eapol_sm_set_event_func(struct eapol_sm *sm, eapol_sm_event_func_t func)
-{
-	sm->event_func = func;
-}
-
 static void eapol_sm_write(struct eapol_sm *sm, const struct eapol_frame *ef,
 				bool noencrypt)
 {
-	__eapol_tx_packet(sm->handshake->ifindex, sm->handshake->aa, ETH_P_PAE,
-				ef, noencrypt);
+	const uint8_t *dst = sm->handshake->authenticator ?
+		sm->handshake->spa : sm->handshake->aa;
+
+	__eapol_tx_packet(sm->handshake->ifindex, dst, ETH_P_PAE, ef,
+				noencrypt);
 }
 
 static inline void handshake_failed(struct eapol_sm *sm, uint16_t reason_code)
 {
-	handshake_event(sm->handshake, HANDSHAKE_EVENT_FAILED, &reason_code);
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_FAILED, reason_code);
 
 	eapol_sm_free(sm);
 }
@@ -950,7 +946,7 @@ static void eapol_install_gtk(struct eapol_sm *sm, uint8_t gtk_key_index,
 	sm->installed_gtk_len = gtk_len;
 }
 
-static void eapol_install_igtk(struct eapol_sm *sm, uint8_t igtk_key_index,
+static void eapol_install_igtk(struct eapol_sm *sm, uint16_t igtk_key_index,
 					const uint8_t *igtk, size_t igtk_len)
 {
 	/*
@@ -967,25 +963,32 @@ static void eapol_install_igtk(struct eapol_sm *sm, uint8_t igtk_key_index,
 	sm->installed_igtk_len = igtk_len - 6;
 }
 
-static void send_eapol_start(struct l_timeout *timeout, void *user_data)
+static void __send_eapol_start(struct eapol_sm *sm, bool noencrypt)
 {
-	struct eapol_sm *sm = user_data;
 	uint8_t buf[sizeof(struct eapol_frame)];
 	struct eapol_frame *frame = (struct eapol_frame *) buf;
 
-	handshake_event(sm->handshake, HANDSHAKE_EVENT_STARTED, NULL);
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_STARTED);
+
+	frame->header.protocol_version = EAPOL_PROTOCOL_VERSION_2001;
+	frame->header.packet_type = 1;
+	l_put_be16(0, &frame->header.packet_len);
+
+	eapol_sm_write(sm, frame, noencrypt);
+}
+
+static void send_eapol_start(struct l_timeout *timeout, void *user_data)
+{
+	struct eapol_sm *sm = user_data;
 
 	l_timeout_remove(sm->eapol_start_timeout);
 	sm->eapol_start_timeout = NULL;
 
-	if (!sm->protocol_version)
-		sm->protocol_version = EAPOL_PROTOCOL_VERSION_2001;
-
-	frame->header.protocol_version = sm->protocol_version;
-	frame->header.packet_type = 1;
-	l_put_be16(0, &frame->header.packet_len);
-
-	eapol_sm_write(sm, frame, false);
+	/*
+	 * AP is probably waiting for us to start, so always send unencrypted
+	 * since the key hasn't been established yet
+	 */
+	__send_eapol_start(sm, true);
 }
 
 static void eapol_set_key_timeout(struct eapol_sm *sm,
@@ -1024,7 +1027,6 @@ static void eapol_set_key_timeout(struct eapol_sm *sm,
 /* 802.11-2016 Section 12.7.6.2 */
 static void eapol_send_ptk_1_of_4(struct eapol_sm *sm)
 {
-	uint32_t ifindex = sm->handshake->ifindex;
 	const uint8_t *aa = sm->handshake->aa;
 	uint8_t frame_buf[512];
 	struct eapol_key *ek = (struct eapol_key *) frame_buf;
@@ -1062,8 +1064,7 @@ static void eapol_send_ptk_1_of_4(struct eapol_sm *sm)
 	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
 			sm->frame_retry);
 
-	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
-				(struct eapol_frame *) ek, false);
+	eapol_sm_write(sm, (struct eapol_frame *) ek, false);
 }
 
 static void eapol_ptk_1_of_4_retry(struct l_timeout *timeout, void *user_data)
@@ -1081,12 +1082,13 @@ static void eapol_ptk_1_of_4_retry(struct l_timeout *timeout, void *user_data)
 }
 
 static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
-					const struct eapol_key *ek)
+					const struct eapol_key *ek,
+					bool unencrypted)
 {
 	const uint8_t *kck;
 	struct eapol_key *step2;
 	uint8_t mic[MIC_MAXLEN];
-	uint8_t *ies;
+	uint8_t ies[512];
 	size_t ies_len;
 	const uint8_t *own_ie = sm->handshake->supplicant_ie;
 	const uint8_t *pmkid;
@@ -1100,7 +1102,11 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 	pmkid = handshake_util_find_pmkid_kde(EAPOL_KEY_DATA(ek, sm->mic_len),
 					EAPOL_KEY_DATA_LEN(ek, sm->mic_len));
 
-	ie_parse_rsne_from_data(own_ie, own_ie[1] + 2, &rsn_info);
+	if (!sm->handshake->wpa_ie) {
+		if (ie_parse_rsne_from_data(own_ie, own_ie[1] + 2,
+						&rsn_info) < 0)
+			goto error_unspecified;
+	}
 
 	/*
 	 * Require the PMKID KDE whenever we've sent a list of PMKIDs in
@@ -1141,7 +1147,7 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 			 * try that, otherwise give up.
 			 */
 			if (sm->eap) {
-				send_eapol_start(NULL, sm);
+				__send_eapol_start(sm, unencrypted);
 				return;
 			}
 
@@ -1171,7 +1177,7 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 			 * layers that we need to do a full reauth
 			 */
 			handshake_event(sm->handshake,
-					HANDSHAKE_EVENT_REKEY_FAILED, NULL);
+					HANDSHAKE_EVENT_REKEY_FAILED);
 			return;
 		}
 
@@ -1193,8 +1199,6 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 		 * Rebuild the RSNE to include the PMKR1Name and append
 		 * MDE + FTE.
 		 */
-		ies = alloca(512);
-
 		rsn_info.num_pmkids = 1;
 		rsn_info.pmkids = sm->handshake->pmk_r1_name;
 
@@ -1208,7 +1212,16 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 		ies_len += fte[1] + 2;
 	} else {
 		ies_len = own_ie[1] + 2;
-		ies = (uint8_t *) own_ie;
+		memcpy(ies, own_ie, ies_len);
+	}
+
+	if (sm->handshake->support_ip_allocation) {
+		/* Wi-Fi P2P Technical Specification v1.7 Table 58 */
+		ies[ies_len++] = IE_TYPE_VENDOR_SPECIFIC;
+		ies[ies_len++] = 4 + 1;
+		l_put_be32(HANDSHAKE_KDE_IP_ADDRESS_REQ, ies + ies_len);
+		ies_len += 4;
+		ies[ies_len++] = 0x01;
 	}
 
 	step2 = eapol_create_ptk_2_of_4(sm->protocol_version,
@@ -1243,7 +1256,7 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 		}
 	}
 
-	eapol_sm_write(sm, (struct eapol_frame *) step2, false);
+	eapol_sm_write(sm, (struct eapol_frame *) step2, unencrypted);
 	l_free(step2);
 
 	l_timeout_remove(sm->eapol_start_timeout);
@@ -1260,19 +1273,17 @@ error_unspecified:
 /* 802.11-2016 Section 12.7.6.4 */
 static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
 {
-	uint32_t ifindex = sm->handshake->ifindex;
 	uint8_t frame_buf[512];
-	uint8_t key_data_buf[128];
+	unsigned int rsne_len = sm->handshake->authenticator_ie[1] + 2;
+	uint8_t key_data_buf[128 + rsne_len];
+	int key_data_len = rsne_len;
 	struct eapol_key *ek = (struct eapol_key *) frame_buf;
-	int key_data_len;
 	enum crypto_cipher cipher = ie_rsn_cipher_suite_to_cipher(
 				sm->handshake->pairwise_cipher);
 	enum crypto_cipher group_cipher = ie_rsn_cipher_suite_to_cipher(
 				sm->handshake->group_cipher);
 	const uint8_t *kck;
 	const uint8_t *kek;
-	struct ie_rsn_info rsn;
-	uint8_t *rsne;
 
 	sm->replay_counter++;
 
@@ -1299,17 +1310,7 @@ static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
 	 * Just one RSNE in Key Data as we only set one cipher in ap->ciphers
 	 * currently.
 	 */
-
-	memset(&rsn, 0, sizeof(rsn));
-	rsn.akm_suites = IE_RSN_AKM_SUITE_PSK;
-	rsn.pairwise_ciphers = sm->handshake->pairwise_cipher;
-	rsn.group_cipher = sm->handshake->group_cipher;
-
-	rsne = key_data_buf;
-	if (!ie_build_rsne(&rsn, rsne))
-		return;
-
-	key_data_len = rsne[1] + 2;
+	memcpy(key_data_buf, sm->handshake->authenticator_ie, rsne_len);
 
 	if (group_cipher) {
 		uint8_t *gtk_kde = key_data_buf + key_data_len;
@@ -1321,6 +1322,21 @@ static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
 		key_data_len += gtk_kde[1] + 2;
 	}
 
+	if (sm->handshake->support_ip_allocation) {
+		/* Wi-Fi P2P Technical Specification v1.7 Table 59 */
+		key_data_buf[key_data_len++] = IE_TYPE_VENDOR_SPECIFIC;
+		key_data_buf[key_data_len++] = 4 + 12;
+		l_put_be32(HANDSHAKE_KDE_IP_ADDRESS_ALLOC,
+				key_data_buf + key_data_len + 0);
+		l_put_be32(sm->handshake->client_ip_addr,
+				key_data_buf + key_data_len + 4);
+		l_put_be32(sm->handshake->subnet_mask,
+				key_data_buf + key_data_len + 8);
+		l_put_be32(sm->handshake->go_ip_addr,
+				key_data_buf + key_data_len + 12);
+		key_data_len += 4 + 12;
+	}
+
 	kek = handshake_state_get_kek(sm->handshake);
 	key_data_len = eapol_encrypt_key_data(kek, key_data_buf,
 						key_data_len, ek, sm->mic_len);
@@ -1341,8 +1357,7 @@ static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
 	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
 			sm->frame_retry);
 
-	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
-				(struct eapol_frame *) ek, false);
+	eapol_sm_write(sm, (struct eapol_frame *) ek, false);
 }
 
 static void eapol_ptk_3_of_4_retry(struct l_timeout *timeout,
@@ -1388,6 +1403,26 @@ static const uint8_t *eapol_find_rsne(const uint8_t *data, size_t data_len,
 	return first;
 }
 
+static const uint8_t *eapol_find_wfa_kde(const uint8_t *data, size_t data_len,
+					uint8_t oi_type)
+{
+	struct ie_tlv_iter iter;
+
+	ie_tlv_iter_init(&iter, data, data_len);
+
+	while (ie_tlv_iter_next(&iter)) {
+		if (ie_tlv_iter_get_tag(&iter) == IE_TYPE_VENDOR_SPECIFIC) {
+			if (!is_ie_wfa_ie(iter.data, iter.len, oi_type))
+				continue;
+		} else
+			continue;
+
+		return ie_tlv_iter_get_data(&iter) - 2;
+	}
+
+	return NULL;
+}
+
 /* 802.11-2016 Section 12.7.6.3 */
 static void eapol_handle_ptk_2_of_4(struct eapol_sm *sm,
 					const struct eapol_key *ek)
@@ -1430,15 +1465,29 @@ static void eapol_handle_ptk_2_of_4(struct eapol_sm *sm,
 	if (!rsne || rsne[1] != sm->handshake->supplicant_ie[1] ||
 			memcmp(rsne + 2, sm->handshake->supplicant_ie + 2,
 				rsne[1])) {
-
 		handshake_failed(sm, MMPDU_REASON_CODE_IE_DIFFERENT);
 		return;
 	}
 
+	if (sm->handshake->support_ip_allocation) {
+		const uint8_t *ip_req_kde =
+			eapol_find_wfa_kde(EAPOL_KEY_DATA(ek, sm->mic_len),
+					EAPOL_KEY_DATA_LEN(ek, sm->mic_len),
+					HANDSHAKE_KDE_IP_ADDRESS_REQ & 255);
+
+		if (ip_req_kde &&
+				(ip_req_kde[1] < 5 || ip_req_kde[6] != 0x01)) {
+			l_debug("Invalid IP Address Request KDE in frame 2/4");
+			handshake_failed(sm, MMPDU_REASON_CODE_INVALID_IE);
+			return;
+		}
+
+		sm->handshake->support_ip_allocation = ip_req_kde != NULL;
+	}
+
 	memcpy(sm->handshake->snonce, ek->key_nonce,
 			sizeof(sm->handshake->snonce));
 	sm->handshake->have_snonce = true;
-	sm->handshake->ptk_complete = true;
 
 	sm->frame_retry = 0;
 
@@ -1463,10 +1512,33 @@ static const uint8_t *eapol_find_wpa_ie(const uint8_t *data, size_t data_len)
 	return NULL;
 }
 
+static bool eapol_check_ip_mask(const uint8_t *mask,
+				const uint8_t *ip1, const uint8_t *ip2)
+{
+	uint32_t mask_uint = l_get_be32(mask);
+	uint32_t ip1_uint = l_get_be32(ip1);
+	uint32_t ip2_uint = l_get_be32(ip2);
+
+	return
+		/* Check IPs are in the same subnet */
+		((ip1_uint ^ ip2_uint) & mask_uint) == 0 &&
+		/* Check IPs are different */
+		ip1_uint != ip2_uint &&
+		/* Check IPs are not subnet addresses */
+		(ip1_uint & ~mask_uint) != 0 &&
+		(ip2_uint & ~mask_uint) != 0 &&
+		/* Check IPs are not broadcast addresses */
+		(ip1_uint | mask_uint) != 0xffffffff &&
+		(ip2_uint | mask_uint) != 0xffffffff &&
+		/* Check the 1s are at the start of the mask */
+		(uint32_t) (mask_uint << __builtin_popcountl(mask_uint)) == 0;
+}
+
 static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 					const struct eapol_key *ek,
 					const uint8_t *decrypted_key_data,
-					size_t decrypted_key_data_size)
+					size_t decrypted_key_data_size,
+					bool unencrypted)
 {
 	const uint8_t *kck;
 	const uint8_t *kek;
@@ -1479,7 +1551,7 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 	const uint8_t *rsne;
 	const uint8_t *optional_rsne = NULL;
 	uint8_t gtk_key_index;
-	uint8_t igtk_key_index;
+	uint16_t igtk_key_index;
 
 	l_debug("ifindex=%u", sm->handshake->ifindex);
 
@@ -1504,13 +1576,17 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 	 * not identical to that the STA received in the Beacon or Probe
 	 * Response frame, the STA shall disassociate.
 	 */
-	if (!sm->handshake->wpa_ie)
+	if (sm->handshake->wpa_ie)
+		rsne = eapol_find_wpa_ie(decrypted_key_data,
+					decrypted_key_data_size);
+	else if (sm->handshake->osen_ie)
+		rsne = eapol_find_wfa_kde(decrypted_key_data,
+					decrypted_key_data_size,
+					IE_WFA_OI_OSEN);
+	else
 		rsne = eapol_find_rsne(decrypted_key_data,
 					decrypted_key_data_size,
 					&optional_rsne);
-	else
-		rsne = eapol_find_wpa_ie(decrypted_key_data,
-					decrypted_key_data_size);
 
 	if (!rsne)
 		goto error_ie_different;
@@ -1528,7 +1604,8 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 		const uint8_t *mde = sm->handshake->mde;
 		const uint8_t *fte = sm->handshake->fte;
 
-		ie_parse_rsne_from_data(rsne, rsne[1] + 2, &ie_info);
+		if (ie_parse_rsne_from_data(rsne, rsne[1] + 2, &ie_info) < 0)
+			goto error_ie_different;
 
 		if (ie_info.num_pmkids != 1 || memcmp(ie_info.pmkids,
 						sm->handshake->pmk_r1_name, 16))
@@ -1587,7 +1664,7 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 		 * pairwise cipher suite which is advertised by an AP, but
 		 * which policy disallows for this particular STA. An
 		 * Authenticator may, therefore, insert a second RSNE to
-		 * overrule the STA’s selection. An Authenticator’s SME shall
+		 * overrule the STA's selection. An Authenticator’s SME shall
 		 * insert the second RSNE, after the first RSNE, only for this
 		 * purpose. The pairwise cipher suite in the second RSNE
 		 * included shall be one of the ciphers advertised by the
@@ -1645,12 +1722,41 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 			return;
 		}
 
-		igtk_key_index = l_get_le16(igtk);;
+		igtk_key_index = l_get_le16(igtk);
 		igtk += 2;
 		igtk_len -= 2;
 	} else
 		igtk = NULL;
 
+	if (sm->handshake->support_ip_allocation) {
+		const uint8_t *ip_alloc_kde =
+			eapol_find_wfa_kde(decrypted_key_data,
+					decrypted_key_data_size,
+					HANDSHAKE_KDE_IP_ADDRESS_ALLOC & 255);
+
+		if (ip_alloc_kde &&
+				(ip_alloc_kde[1] < 16 ||
+				 !eapol_check_ip_mask(ip_alloc_kde + 10,
+							ip_alloc_kde + 6,
+							ip_alloc_kde + 14))) {
+			l_debug("Invalid IP Allocation KDE in frame 3/4");
+			handshake_failed(sm, MMPDU_REASON_CODE_INVALID_IE);
+			return;
+		}
+
+		sm->handshake->support_ip_allocation = ip_alloc_kde != NULL;
+
+		if (ip_alloc_kde) {
+			sm->handshake->client_ip_addr =
+				l_get_be32(ip_alloc_kde + 6);
+			sm->handshake->subnet_mask =
+				l_get_be32(ip_alloc_kde + 10);
+			sm->handshake->go_ip_addr =
+				l_get_be32(ip_alloc_kde + 14);
+		} else
+			l_debug("Authenticator ignored our IP Address Request");
+	}
+
 retransmit:
 	/*
 	 * 802.11-2016, Section 12.7.6.4:
@@ -1694,7 +1800,7 @@ retransmit:
 		}
 	}
 
-	eapol_sm_write(sm, (struct eapol_frame *) step4, false);
+	eapol_sm_write(sm, (struct eapol_frame *) step4, unencrypted);
 	l_free(step4);
 
 	if (sm->handshake->ptk_complete)
@@ -1753,13 +1859,22 @@ static void eapol_handle_ptk_4_of_4(struct eapol_sm *sm,
 	l_timeout_remove(sm->timeout);
 	sm->timeout = NULL;
 
-	handshake_state_install_ptk(sm->handshake);
+	/*
+	 * If ptk_complete is set, then we are receiving Message 4 again.
+	 * This might be a retransmission, so accept but don't install
+	 * the keys again.
+	 */
+	if (!sm->handshake->ptk_complete)
+		handshake_state_install_ptk(sm->handshake);
+
+	sm->handshake->ptk_complete = true;
 }
 
 static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 					const struct eapol_key *ek,
 					const uint8_t *decrypted_key_data,
-					size_t decrypted_key_data_size)
+					size_t decrypted_key_data_size,
+					bool unencrypted)
 {
 	const uint8_t *kck;
 	struct eapol_key *step2;
@@ -1769,7 +1884,7 @@ static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 	uint8_t gtk_key_index;
 	const uint8_t *igtk;
 	size_t igtk_len;
-	uint8_t igtk_key_index;
+	uint16_t igtk_key_index;
 
 	if (!eapol_verify_gtk_1_of_2(ek, sm->handshake->wpa_ie, sm->mic_len)) {
 		handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
@@ -1804,7 +1919,7 @@ static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 		if (!igtk)
 			return;
 
-		igtk_key_index = l_get_le16(igtk);;
+		igtk_key_index = l_get_le16(igtk);
 		igtk += 2;
 		igtk_len -= 2;
 	} else
@@ -1858,7 +1973,7 @@ static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 		}
 	}
 
-	eapol_sm_write(sm, (struct eapol_frame *) step2, false);
+	eapol_sm_write(sm, (struct eapol_frame *) step2, unencrypted);
 	l_free(step2);
 
 	eapol_install_gtk(sm, gtk_key_index, gtk, gtk_len, ek->key_rsc);
@@ -1889,7 +2004,8 @@ static struct eapol_sm *eapol_find_sm(uint32_t ifindex, const uint8_t *aa)
 }
 
 static void eapol_key_handle(struct eapol_sm *sm,
-				const struct eapol_frame *frame)
+				const struct eapol_frame *frame,
+				bool unencrypted)
 {
 	const struct eapol_key *ek;
 	const uint8_t *kck;
@@ -1990,13 +2106,13 @@ static void eapol_key_handle(struct eapol_sm *sm,
 			goto done;
 
 		eapol_handle_gtk_1_of_2(sm, ek, decrypted_key_data,
-					key_data_len);
+					key_data_len, unencrypted);
 		goto done;
 	}
 
 	/* If no MIC, then assume packet 1, otherwise packet 3 */
 	if (!ek->key_mic && !ek->encrypted_key_data)
-		eapol_handle_ptk_1_of_4(sm, ek);
+		eapol_handle_ptk_1_of_4(sm, ek, unencrypted);
 	else {
 		if (!key_data_len)
 			goto done;
@@ -2004,7 +2120,7 @@ static void eapol_key_handle(struct eapol_sm *sm,
 		eapol_handle_ptk_3_of_4(sm, ek,
 					decrypted_key_data ?:
 					EAPOL_KEY_DATA(ek, sm->mic_len),
-					key_data_len);
+					key_data_len, unencrypted);
 	}
 
 done:
@@ -2028,7 +2144,7 @@ static void eapol_eap_msg_cb(const uint8_t *eap_data, size_t len,
 
 	memcpy(frame->data, eap_data, len);
 
-	eapol_sm_write(sm, frame, false);
+	eapol_sm_write(sm, frame, sm->last_eap_unencrypted);
 }
 
 /* This respresentes the eapTimout, eapFail and eapSuccess messages */
@@ -2048,6 +2164,18 @@ static void eapol_eap_complete_cb(enum eap_result result, void *user_data)
 	}
 
 	eap_reset(sm->eap);
+
+	if (sm->handshake->authenticator) {
+		if (L_WARN_ON(!sm->handshake->have_pmk)) {
+			handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
+			return;
+		}
+
+		/* sm->mic_len will have been set in eapol_eap_results_cb */
+
+		/* Kick off 4-Way Handshake */
+		eapol_ptk_1_of_4_retry(NULL, sm);
+	}
 }
 
 /* This respresentes the eapResults message */
@@ -2131,10 +2259,8 @@ static void eapol_eap_event_cb(unsigned int event,
 {
 	struct eapol_sm *sm = user_data;
 
-	if (!sm->event_func)
-		return;
-
-	sm->event_func(event, event_data, sm->user_data);
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_EAP_NOTIFY, event,
+			event_data);
 }
 
 void eapol_sm_set_use_eapol_start(struct eapol_sm *sm, bool enabled)
@@ -2156,6 +2282,7 @@ static void eapol_auth_key_handle(struct eapol_sm *sm,
 	size_t frame_len = 4 + L_BE16_TO_CPU(frame->header.packet_len);
 	const struct eapol_key *ek = eapol_key_validate((const void *) frame,
 							frame_len, sm->mic_len);
+	uint16_t key_data_len;
 
 	if (!ek)
 		return;
@@ -2170,7 +2297,8 @@ static void eapol_auth_key_handle(struct eapol_sm *sm,
 	if (!sm->handshake->have_anonce)
 		return; /* Not expecting an EAPoL-Key yet */
 
-	if (!sm->handshake->ptk_complete)
+	key_data_len = EAPOL_KEY_DATA_LEN(ek, sm->mic_len);
+	if (key_data_len != 0)
 		eapol_handle_ptk_2_of_4(sm, ek);
 	else
 		eapol_handle_ptk_4_of_4(sm, ek);
@@ -2178,14 +2306,43 @@ static void eapol_auth_key_handle(struct eapol_sm *sm,
 
 static void eapol_rx_auth_packet(uint16_t proto, const uint8_t *from,
 				const struct eapol_frame *frame,
+				bool noencrypt,
 				void *user_data)
 {
 	struct eapol_sm *sm = user_data;
 
-	if (!sm->protocol_version)
-		sm->protocol_version = frame->header.protocol_version;
+	if (proto != ETH_P_PAE || memcmp(from, sm->handshake->spa, 6))
+		return;
 
 	switch (frame->header.packet_type) {
+	case 0:	/* EAPOL-EAP */
+		if (!sm->eap) {
+			l_error("Authenticator received an unexpected "
+				"EAPOL-EAP frame from %s",
+				util_address_to_string(from));
+			return;
+		}
+
+		eap_rx_packet(sm->eap, frame->data,
+				L_BE16_TO_CPU(frame->header.packet_len));
+		break;
+
+	case 1:	/* EAPOL-Start */
+		/*
+		 * The supplicant may have sent an EAPoL-Start even before
+		 * we queued our EAP Identity Request or it may have missed our
+		 * early Identity Request and may need a retransmission.  Tell
+		 * sm->eap so it can decide whether to send a new Identity
+		 * Request or ignore this.
+		 *
+		 * TODO: if we're already past the full handshake, send a
+		 * new msg 1/4.
+		 */
+		if (sm->eap)
+			eap_start(sm->eap);
+
+		break;
+
 	case 3: /* EAPOL-Key */
 		eapol_auth_key_handle(sm, frame);
 		break;
@@ -2200,6 +2357,7 @@ static void eapol_rx_auth_packet(uint16_t proto, const uint8_t *from,
 
 static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 				const struct eapol_frame *frame,
+				bool unencrypted,
 				void *user_data)
 {
 	struct eapol_sm *sm = user_data;
@@ -2219,6 +2377,7 @@ static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 			return;
 
 		sm->early_frame = l_memdup(frame, len);
+		sm->early_frame_unencrypted = unencrypted;
 
 		return;
 	}
@@ -2244,6 +2403,7 @@ static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 		}
 
 		sm->eap_exchanged = true;
+		sm->last_eap_unencrypted = unencrypted;
 
 		eap_rx_packet(sm->eap, frame->data,
 				L_BE16_TO_CPU(frame->header.packet_len));
@@ -2261,13 +2421,16 @@ static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 			 * use a cached PMK.  We don't yet cache PMKs so
 			 * send an EAPOL-Start if we haven't sent one yet.
 			 */
-			if (sm->eapol_start_timeout)
-				send_eapol_start(NULL, sm);
+			if (sm->eapol_start_timeout) {
+				l_timeout_remove(sm->eapol_start_timeout);
+				sm->eapol_start_timeout = NULL;
+				__send_eapol_start(sm, unencrypted);
+			}
 
 			return;
 		}
 
-		eapol_key_handle(sm, frame);
+		eapol_key_handle(sm, frame, unencrypted);
 		break;
 
 	default:
@@ -2308,21 +2471,14 @@ void __eapol_set_rekey_offload_func(eapol_rekey_offload_func_t func)
 
 void eapol_register(struct eapol_sm *sm)
 {
+	eapol_frame_watch_func_t rx_handler = sm->handshake->authenticator ?
+		eapol_rx_auth_packet : eapol_rx_packet;
+
 	l_queue_push_head(state_machines, sm);
 
-	if (sm->handshake->authenticator) {
-		sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
-						eapol_rx_auth_packet, sm);
-
-		sm->started = true;
-		/* Since AP/AdHoc only support AKM PSK we can hard code this */
-		sm->mic_len = 16;
-
-		/* kick off handshake */
-		eapol_ptk_1_of_4_retry(NULL, sm);
-	} else
-		sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
-						eapol_rx_packet, sm);
+	sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
+						rx_handler, sm);
+	sm->protocol_version = sm->handshake->proto_version;
 }
 
 bool eapol_start(struct eapol_sm *sm)
@@ -2347,19 +2503,11 @@ bool eapol_start(struct eapol_sm *sm)
 
 	sm->started = true;
 
-	/*
-	 * FILS only uses the 4-way for rekeys, so only started needs to be set,
-	 * then we wait for a rekey.
-	 */
-	if (sm->handshake->akm_suite & (IE_RSN_AKM_SUITE_FILS_SHA256 |
-			IE_RSN_AKM_SUITE_FILS_SHA384))
-		return true;
-
 	if (sm->require_handshake)
 		sm->timeout = l_timeout_create(eapol_4way_handshake_time,
 				eapol_timeout, sm, NULL);
 
-	if (sm->use_eapol_start) {
+	if (!sm->handshake->authenticator && sm->use_eapol_start) {
 		/*
 		 * We start a short timeout, if EAP packets are not received
 		 * from AP, then we send the EAPoL-Start
@@ -2374,11 +2522,32 @@ bool eapol_start(struct eapol_sm *sm)
 	/* Process any frames received early due to scheduling */
 	if (sm->early_frame) {
 		eapol_rx_packet(ETH_P_PAE, sm->handshake->aa,
-				sm->early_frame, sm);
+				sm->early_frame, sm->early_frame_unencrypted,
+				sm);
 		l_free(sm->early_frame);
 		sm->early_frame = NULL;
 	}
 
+	if (sm->handshake->authenticator) {
+		if (!sm->protocol_version)
+			sm->protocol_version = EAPOL_PROTOCOL_VERSION_2004;
+
+		if (sm->handshake->settings_8021x) {
+			/*
+			 * If we're allowed to, send EAP Identity request
+			 * immediately, otherwise wait for an EAPoL-Start.
+			 */
+			if (!sm->use_eapol_start)
+				eap_start(sm->eap);
+		} else {
+			if (L_WARN_ON(!sm->handshake->have_pmk))
+				return false;
+
+			/* Kick off handshake */
+			eapol_ptk_1_of_4_retry(NULL, sm);
+		}
+	}
+
 	return true;
 
 eap_error:
@@ -2435,6 +2604,7 @@ static void preauth_frame(struct preauth_sm *sm, uint8_t packet_type,
 
 static void preauth_rx_packet(uint16_t proto, const uint8_t *from,
 				const struct eapol_frame *frame,
+				bool unencrypted,
 				void *user_data)
 {
 	struct preauth_sm *sm = user_data;
@@ -2442,6 +2612,15 @@ static void preauth_rx_packet(uint16_t proto, const uint8_t *from,
 	if (proto != 0x88c7 || memcmp(from, sm->aa, 6))
 		return;
 
+	/*
+	 * We do not expect any pre-auth packets to be unencrypted
+	 * since we're authenticating via the currently connected AP
+	 * and pre-authentication implies we are already connected
+	 * and the keys are set
+	 */
+	if (L_WARN_ON(unencrypted))
+		return;
+
 	if (frame->header.packet_type != 0) /* EAPOL-EAP */
 		return;
 
@@ -2628,7 +2807,8 @@ void __eapol_rx_packet(uint32_t ifindex, const uint8_t *src, uint16_t proto,
 					eapol_frame_watch_match_ifindex,
 					L_UINT_TO_PTR(ifindex),
 					eapol_frame_watch_func_t, proto, src,
-					(const struct eapol_frame *) eh);
+					(const struct eapol_frame *) eh,
+					noencrypt);
 }
 
 void __eapol_tx_packet(uint32_t ifindex, const uint8_t *dst, uint16_t proto,
@@ -2643,20 +2823,20 @@ void __eapol_tx_packet(uint32_t ifindex, const uint8_t *dst, uint16_t proto,
 void __eapol_set_config(struct l_settings *config)
 {
 	if (!l_settings_get_uint(config, "EAPoL",
-			"max_4way_handshake_time", &eapol_4way_handshake_time))
+			"MaxHandshakeTime", &eapol_4way_handshake_time))
 		eapol_4way_handshake_time = 5;
 }
 
-bool eapol_init()
+int eapol_init(void)
 {
 	state_machines = l_queue_new();
 	preauths = l_queue_new();
 	watchlist_init(&frame_watches, &eapol_frame_watch_ops);
 
-	return true;
+	return 0;
 }
 
-bool eapol_exit()
+void eapol_exit(void)
 {
 	if (!l_queue_isempty(state_machines))
 		l_warn("stale eapol state machines found");
@@ -2669,6 +2849,6 @@ bool eapol_exit()
 	l_queue_destroy(preauths, preauth_sm_destroy);
 
 	watchlist_destroy(&frame_watches);
-
-	return true;
 }
+
+IWD_MODULE(eapol, eapol_init, eapol_exit);
